"""Testing events splitter

SPDX-FileCopyrightText: 2024 CERT.at GmbH <https://cert.at/>
SPDX-License-Identifier: AGPL-3.0-or-later
"""

import json
import unittest

from intelmq_extensions.bots.experts.event_group_splitter.expert import (
    EventGroupSplitterExpertBot,
)

from ....base import BotTestCase

INPUT = {
    "__type": "Event",
    "classification.identifier": "zeus",
    "classification.type": "infected-system",
    "notify": False,
    "source.asn": 1,
    "source.ip": "192.0.2.1",
    "feed.name": "Example Feed",
    "feed.code": "feed-1",
}


class TestEventGroupSplitter(BotTestCase, unittest.TestCase):
    "Test cases prepared to handle tagging in ShadowServer Compromised Website report"

    @classmethod
    def set_bot(cls):
        cls.bot_reference = EventGroupSplitterExpertBot
        cls.sysconfig = {
            "look_in": "extra.tag",
            "copy_to": ["classification.identifier"],
            "regex": r"([a-zA-Z0-9\-]+)[;]{0,1}",
            "ignore": ["tag2", "tag4"],
            "groups_file": None,
        }

    def setUp(self) -> None:
        super().setUp()
        self.create_dynamic_groups_file()

    def create_dynamic_groups_file(self):
        self.groups_file_path = f"{self.tmp_dir}/groups.json"
        groups = {
            "feed-1": [
                ["citrix", "injected-code"],
                ["backdoor-activity", "ivanti-connect-secure"],
                ["ivanti-connect-secure", "credential-stealer", "injected-code"],
            ],
            "feed-3": [["test"]],
        }
        with open(self.groups_file_path, "w+") as f:
            json.dump(groups, f)

    def test_ignore_selected_tags(self):
        message = {
            **INPUT,
            "extra.tag": "tag1,tag2,tag3,tag4",
        }
        self.input_message = message

        self.run_bot(parameters={"groups_file": self.groups_file_path})

        self.assertOutputQueueLen(1)
        self.assertMessageEqual(
            0,
            {
                **message,
                "classification.identifier": "tag1-tag3",
            },
        )

    def test_handle_configured_groups(self):
        message = {
            **INPUT,
            "extra.tag": "citrix;injected-code;backdoor-activity;ivanti-connect-secure",
        }
        self.input_message = message

        self.run_bot(parameters={"groups_file": self.groups_file_path})

        self.assertOutputQueueLen(2)
        self.assertMessageEqual(
            0,
            {
                **message,
                "classification.identifier": "citrix-injected-code",
            },
        )
        self.assertMessageEqual(
            1,
            {
                **message,
                "classification.identifier": "backdoor-activity-ivanti-connect-secure",
            },
        )

    def test_generate_new_groups(self):
        message = {
            **INPUT,
            "extra.tag": "injected-code;cisco;backdoor-activity;ivanti-connect-secure",
        }
        self.input_message = message

        self.run_bot(parameters={"groups_file": self.groups_file_path})

        self.assertOutputQueueLen(2)
        self.assertMessageEqual(
            0,
            {
                **message,
                "classification.identifier": "backdoor-activity-ivanti-connect-secure",
            },
        )
        self.assertMessageEqual(
            1,
            {
                **message,
                "classification.identifier": "cisco-injected-code",
            },
        )

    def test_generate_new_groups_when_duplicates(self):
        message = {
            **INPUT,
            "extra.tag": "citrix;injected-code;injected-code;cisco",
        }
        self.input_message = message

        self.run_bot(parameters={"groups_file": self.groups_file_path})

        self.assertOutputQueueLen(2)
        self.assertMessageEqual(
            0,
            {
                **message,
                "classification.identifier": "citrix-injected-code",
            },
        )
        self.assertMessageEqual(
            1,
            {
                **message,
                "classification.identifier": "cisco-injected-code",
            },
        )

        with open(self.groups_file_path) as f:
            new_groups = json.load(f)["feed-1"]
        self.assertEqual(4, len(new_groups))
        self.assertIn(set(["cisco", "injected-code"]), (set(g) for g in new_groups))

    def test_new_groups_are_kept(self):
        message = {
            **INPUT,
            "extra.tag": "injected-code;cisco",
        }
        self.input_message = message

        self.run_bot(parameters={"groups_file": self.groups_file_path})

        self.assertOutputQueueLen(1)
        self.assertMessageEqual(
            0,
            {
                **message,
                "classification.identifier": "cisco-injected-code",
            },
        )

        with open(self.groups_file_path) as f:
            new_groups = json.load(f)["feed-1"]
        self.assertIn(set(["cisco", "injected-code"]), (set(g) for g in new_groups))

        message = {
            **INPUT,
            "extra.tag": "cisco;injected-code;new-tag",
        }
        self.input_message = message

        self.run_bot(parameters={"groups_file": self.groups_file_path})
        self.assertOutputQueueLen(2)
        self.assertMessageEqual(
            0,
            {
                **message,
                "classification.identifier": "cisco-injected-code",
            },
        )
        self.assertMessageEqual(
            1,
            {
                **message,
                "classification.identifier": "new-tag",
            },
        )

        with open(self.groups_file_path) as f:
            new_groups = json.load(f)["feed-1"]
        self.assertEqual(5, len(new_groups))
        self.assertIn(set(["cisco", "injected-code"]), (set(g) for g in new_groups))
        self.assertIn(set(["new-tag"]), (set(g) for g in new_groups))

    def test_groups_are_separated_between_feeds(self):
        message = {
            **INPUT,
            "extra.tag": "citrix,injected-code",
            "feed.code": "feed-2",
        }
        self.input_message = message

        self.run_bot(parameters={"groups_file": self.groups_file_path})

        self.assertOutputQueueLen(1)
        self.assertMessageEqual(
            0,
            {
                **message,
                "classification.identifier": "citrix-injected-code",
            },
        )

        with open(self.groups_file_path) as f:
            new_groups = json.load(f)
        self.assertEqual(3, len(new_groups["feed-1"]))
        self.assertEqual(1, len(new_groups["feed-2"]))
        self.assertIn(
            set(["citrix", "injected-code"]), (set(g) for g in new_groups["feed-2"])
        )

    def test_groups_are_not_separated_between_feeds_if_set(self):
        message = {
            **INPUT,
            "extra.tag": "citrix,injected-code",
            "feed.code": "feed-2",
        }
        self.input_message = message

        self.prepare_bot(
            parameters={
                "groups_file": self.groups_file_path,
                "treat_feeds_separately": False,
            },
            destination_queues=["_default", "new_tag_groups"],
        )
        self.run_bot(prepare=False)

        self.assertOutputQueueLen(1)
        self.assertOutputQueueLen(0, "new_tag_groups")
        self.assertMessageEqual(
            0,
            {
                **message,
                "classification.identifier": "citrix-injected-code",
            },
        )

        with open(self.groups_file_path) as f:
            new_groups = json.load(f)
        self.assertEqual(3, len(new_groups["feed-1"]))
        self.assertNotIn("feed-2", new_groups)

    def test_events_with_new_tag_groups_are_sent_to_specific_path(self):
        message_with_old_tag = {**INPUT, "extra.tag": "citrix,injected-code"}
        self.input_message = message_with_old_tag

        self.prepare_bot(
            parameters={"groups_file": self.groups_file_path},
            destination_queues=["_default", "new_tag_groups"],
        )
        self.run_bot(prepare=False)

        self.assertOutputQueueLen(1)
        self.assertOutputQueueLen(0, "new_tag_groups")

        message_with_new_tags = {**INPUT, "extra.tag": "new,group"}
        self.input_message = message_with_new_tags

        self.prepare_bot(
            parameters={"groups_file": self.groups_file_path},
            destination_queues=["_default", "new_tag_groups"],
        )
        self.run_bot(prepare=False)

        self.assertOutputQueueLen(1)
        self.assertOutputQueueLen(1, "new_tag_groups")

        # new_tag_groups should be notified only once
        self.input_message = message_with_new_tags

        self.prepare_bot(
            parameters={"groups_file": self.groups_file_path},
            destination_queues=["_default", "new_tag_groups"],
        )
        self.run_bot(prepare=False)

        self.assertOutputQueueLen(1)
        self.assertOutputQueueLen(0, "new_tag_groups")



# -*- coding: utf-8 -*-
"""
CERT.at geolocate the national CERT abuse service
"""
from intelmq.lib.bot import Bot

try:
    import psycopg2
except ImportError:
    psycopg2 = None


class CERTatContactExpertBot(Bot):
    connect_timeout = 5
    database: str = ""
    user: str = ""
    password: str = ""
    host: str = ""
    port: str = ""
    sslmode: str = ""
    autocommit: bool = True
    table: str = "contacts"
    column: str = ""
    feed_code: str = ""
    ascolumn: str = ""
    overwrite: bool = False

    def init(self):
        self.logger.debug("Connecting to database.")
        if psycopg2 is None:
            self.logger.error("Could not import psycopg2. Please install it.")
            self.stop()

        try:
            self.con = psycopg2.connect(
                database=self.database,
                user=self.user,
                password=self.password,
                host=self.host,
                port=self.port,
                sslmode=self.sslmode,
                connect_timeout=self.connect_timeout,
            )
            self.cur = self.con.cursor()
            self.con.autocommit = self.autocommit

        except Exception:
            self.logger.exception("Failed to connect to database.")
            self.stop()
        self.logger.info("Connected to PostgreSQL.")

        self.query = (
            'SELECT "{column}", "can-see-tlp-amber_{feed_code}"'
            ' FROM "{table}" WHERE "{ascolumn}" = %s'
            "".format(
                table=self.table,
                column=self.column,
                feed_code=self.feed_code,
                ascolumn=self.ascolumn,
            )
        )

    def process(self):
        event = self.receive_message()
        default_destination_visible = (
            True if event.get("feed.code") != self.feed_code else False
        )

        if "source.asn" not in event:
            self.logger.info("source.asn not present in event. Skipping event.")
            event.add(
                "destination_visible",
                default_destination_visible,
                overwrite=self.overwrite,
            )
            self.send_message(event)
            self.acknowledge_message()
            return

        if "source.abuse_contact" in event and not self.overwrite:
            event.add(
                "destination_visible",
                default_destination_visible,
                overwrite=self.overwrite,
            )
            self.send_message(event)
            self.acknowledge_message()
            return

        try:
            self.logger.debug(
                "Executing %r." % self.cur.mogrify(self.query, (event["source.asn"],))
            )
            self.cur.execute(self.query, (event["source.asn"],))
        except (
            psycopg2.InterfaceError,
            psycopg2.InternalError,
            psycopg2.OperationalError,
            AttributeError,
        ):
            self.logger.exception("Database connection problem, connecting again.")
            self.init()
        else:
            if self.cur.rowcount > 1:
                raise ValueError(
                    "Lookup returned more than one result. Please inspect."
                )
            elif self.cur.rowcount == 1:
                result = self.cur.fetchone()
                self.logger.debug(
                    "Changing `source.abuse_contact` from %r to %r."
                    % (event.get("source.abuse_contact"), result[0])
                )

                event.add("source.abuse_contact", result[0], overwrite=self.overwrite)

                if event["feed.code"] == self.feed_code:
                    if result[1]:
                        event.add("destination_visible", True, overwrite=self.overwrite)
                    else:
                        event.add(
                            "destination_visible", False, overwrite=self.overwrite
                        )
                else:
                    event.add("destination_visible", True, overwrite=self.overwrite)

            else:
                self.logger.debug("No contact found.")
                event.add(
                    "destination_visible",
                    default_destination_visible,
                    overwrite=self.overwrite,
                )

            self.send_message(event)
            self.acknowledge_message()


BOT = CERTatContactExpertBot



import unittest
from unittest import mock

import requests
from requests_mock import MockerCore

from intelmq_extensions.bots.outputs.mattermost.output import MattermostOutputBot

from ....base import BotTestCase

INPUT = {
    "__type": "Event",
    "classification.identifier": "zeus",
    "classification.type": "infected-system",
    "notify": False,
    "source.asn": 1,
    "source.ip": "192.0.2.1",
    "feed.name": "Example Feed",
}


class TestMattermostOutputBot(BotTestCase, unittest.TestCase):
    MM_URL = "https://my.mm.instance"
    BOT_TOKEN = "XYZ"
    CHANNEL = "channel-1"

    @classmethod
    def set_bot(cls):
        cls.bot_reference = MattermostOutputBot
        cls.sysconfig = {
            "mm_url": cls.MM_URL,
            "bot_token": cls.BOT_TOKEN,
            "channel_id": cls.CHANNEL,
            "author_name": None,
            "author_icon": None,
        }

    def mock_http_session(self):
        session = requests.Session()
        session_mock = mock.patch(
            "intelmq_extensions.bots.outputs.mattermost.output.create_request_session",
            return_value=session,
        )
        session_mock.start()
        self.addCleanup(session_mock.stop)

        self.requests = MockerCore(session=session)
        self.requests.start()
        self.addCleanup(self.requests.stop)

    def mock_request(
        self,
        message: str = None,
        attachment: dict = None,
        card: str = None,
        path: str = "api/v4/posts",
        method: str = "post",
        **kwargs,
    ):
        data = {"channel_id": self.CHANNEL}
        if message is not None:
            data["message"] = message
        data["props"] = {}
        if attachment is not None:
            data["props"]["attachments"] = [attachment]
        if card is not None:
            data["props"]["card"] = card
        mocking_method = getattr(self.requests, method)
        mocking_method(
            f"{self.MM_URL}/{path}",
            request_headers={
                "Authorization": f"Bearer {self.BOT_TOKEN}",
            },
            **kwargs,
        )
        return data

    def setUp(self):
        super().setUp()
        self.mock_http_session()

    def test_simple_static_message(self):
        expected_payload = self.mock_request(message="This is a static message")

        self.input_message = INPUT
        self.run_bot(parameters={"message": "This is a static message"})

        self.assertEqual(1, self.requests.call_count)
        self.assertEqual(expected_payload, self.requests.last_request.json())

    def test_get_info_from_event_if_allowed(self):
        expected_payload = self.mock_request(
            message="message zeus",
            attachment={
                "pretext": "pretext zeus",
                "text": "text zeus",
                "fallback": "fallback zeus",
                "title": "title zeus",
                "title_link": "title_link {ev[classification.identifier]}",
                "author_name": "author_name {ev[classification.identifier]}",
                "author_icon": "author_icon {ev[classification.identifier]}",
                "author_link": "author_link {ev[classification.identifier]}",
                "color": "color {ev[classification.identifier]}",
                "footer": "footer zeus",
                "fields": [
                    {"short": True, "title": "field title zeus", "value": "value zeus"}
                ],
            },
            card="card zeus",
        )

        self.input_message = INPUT
        self.run_bot(
            parameters={
                "message": "message {ev[classification.identifier]}",
                "card": "card {ev[classification.identifier]}",
                "pretext": "pretext {ev[classification.identifier]}",
                "text": "text {ev[classification.identifier]}",
                "fallback": "fallback {ev[classification.identifier]}",
                "title": "title {ev[classification.identifier]}",
                "title_link": "title_link {ev[classification.identifier]}",
                "author_name": "author_name {ev[classification.identifier]}",
                "author_icon": "author_icon {ev[classification.identifier]}",
                "author_link": "author_link {ev[classification.identifier]}",
                "color": "color {ev[classification.identifier]}",
                "footer": "footer {ev[classification.identifier]}",
                "fields": [
                    {
                        "short": True,
                        "title": "field title {ev[classification.identifier]}",
                        "value": "value {ev[classification.identifier]}",
                    }
                ],
            }
        )

        self.assertEqual(1, self.requests.call_count)
        self.assertEqual(expected_payload, self.requests.last_request.json())



"""Parsing DISP data feed.

Currently only for credentials tracking

SPDX-FileCopyrightText: 2023 CERT.at GmbH <https://cert.at/>
SPDX-License-Identifier: AGPL-3.0-or-later
"""

import json
from copy import deepcopy
from datetime import datetime
from urllib import parse

import intelmq.lib.message as message
from dateutil.parser import ParserError
from dateutil.parser import parse as dt_parse
from dateutil.tz import UTC
from dns.exception import DNSException
from intelmq.lib import utils
from intelmq.lib.bot import ParserBot
from intelmq.lib.exceptions import IntelMQException
from intelmq.lib.harmonization import URL, DateTime


class InvalidData(IntelMQException, ValueError):
    """Given message is invalid comparing with parser requirements"""


class DISPParserBot(ParserBot):
    compromise_time_format: str = "%Y-%m"  # empty, 'original' or format string
    redact_url_path: bool = True
    resolve_ip: bool = False

    def parse(self, report: message.Report):
        raw_report = utils.base64_decode(report.get("raw"))
        report_data = json.loads(raw_report)

        incident = report_data.get("incident")
        evidences = report_data.get("evidences", {}).get("credentials")

        if not evidences:
            self.logger.error("Report doesn't contain any evidences")
            raise InvalidData("No evidences in DISP report")

        for evidence in evidences:
            current = {"incident": deepcopy(incident), "evidence": evidence}
            self._current_line = json.dumps(current)
            yield current

    def parse_line(self, line: dict, report: message.Report):
        event = self.new_event(report)
        event.add("raw", json.dumps(line))

        incident, evidence = line.get("incident"), line.get("evidence")
        self._map_incident(incident, event)
        self._map_evidence(evidence, event)
        return event

    def _map_incident(self, incident: dict, event: message.Event):
        event.add("event_description.text", incident.get("title"))
        event.add("extra.feed_event_id", incident.get("id"))
        event.add(
            "time.source", DateTime.from_epoch_millis(incident.get("validationDate"))
        )
        event.add("extra.monitored_asset", incident.get("relatedAssets", [""])[0])

    def _map_evidence(self, evidence: dict, event: message.Event):
        url = evidence.get("url")
        parsed_url = parse.urlsplit(url)
        event.add("source.fqdn", parsed_url.netloc)

        if self.resolve_ip:
            try:
                event.add("source.ip", URL.to_ip(url))
            except DNSException:
                self.logger.warning("Cannot get IP for domain %s.", url, exc_info=True)

        path = parsed_url.path
        if self.redact_url_path and parsed_url.path and parsed_url.path != "/":
            path = "/[REDACTED]"
        event.add(
            "source.url",
            f"{parsed_url.scheme}://{parsed_url.netloc}{path}",
            sanitize=True,
        )
        event.add("source.urlpath", path)
        event.add("extra.full_url", self._disarm_url(url))

        event.add("source.account", evidence.get("username"))
        event.add("extra.account", evidence.get("username"))
        event.add("extra.password", evidence.get("password"))
        event.add("extra.application", evidence.get("application"))
        event.add("extra.compromise_time", self._parse_compromise_time(evidence))
        event.add("extra.compromise_time_full", evidence.get("date"))

        event.add("malware.name", evidence.get("malware"))

    @staticmethod
    def _disarm_url(url: str):
        """Prevents URLs being usable"""
        url = url.replace("http://", "hxxp://")
        return url.replace("https://", "hxxps://")

    def _parse_compromise_time(self, evidence: dict):
        compromise_time = evidence.get("date")
        if not compromise_time:
            return None

        if self.compromise_time_format == "original":
            return compromise_time

        if not self.compromise_time_format:
            return ""

        try:
            dt: datetime = dt_parse(compromise_time).astimezone(tz=UTC)
        except ParserError:
            self.logger.warning(
                "Error parsing compromise time %s.", compromise_time, exc_info=True
            )
            return None
        return dt.strftime(self.compromise_time_format)


BOT = DISPParserBot



"""
XMPP Collector Bot
Connects to a XMPP Server and a Room and reads data from the room.
If no room is provided, which is equivalent to an empty string,
it only collects events which were sent to the xmpp user directly.

TLS is used by default.

Tested with Python >= 3.4
Tested with slixmpp >= 1.0.0-beta5

Copyright (C) 2016 by Bundesamt für Sicherheit in der Informationstechnik
Software engineering by Intevation GmbH

Parameters:
ca_certs: string to a CA-bundle file or false/empty string for no checks
strip_message: boolean
xmpp_user: string
xmpp_server: string
xmpp_password: boolean
xmpp_room: string
xmpp_room_password: string
xmpp_room_nick: string
pass_full_xml: boolean
strip_message: boolean
xmpp_userlist: array
xmpp_whitelist_mode: boolean
"""

from intelmq.lib.bot import CollectorBot
from intelmq.lib.exceptions import MissingDependencyError

try:
    import slixmpp

    class XMPPClient(slixmpp.ClientXMPP):
        def __init__(self, jid, password, room, room_nick, room_password, logger):
            slixmpp.ClientXMPP.__init__(self, jid, password)

            self.logger = logger
            self.logger.info("Initiated.")
            self.xmpp_room = room
            self.xmpp_room_nick = room_nick
            self.xmpp_room_password = room_password

            self.add_event_handler("session_start", self.session_start)

        def session_start(self, event):
            self.send_presence()
            self.logger.debug("Session started.")

            try:
                self.get_roster()
            except slixmpp.exceptions.IqError as err:
                self.logger.error("There was an error getting the roster.")
                self.logger.error(err.iq["error"]["condition"])
                self.disconnect()
            except slixmpp.exceptions.IqTimeout:
                self.logger.error("Server is taking too long to respond.")
                self.disconnect()

            if (
                self.xmpp_room
            ):  # and self.plugin.get('xep_0045') # this check should also exist!
                self.logger.debug("Joining room: %s.", self.xmpp_room)
                pwd = self.xmpp_room_password if self.xmpp_room_password else ""
                self.plugin["xep_0045"].joinMUC(
                    self.xmpp_room, self.xmpp_room_nick, password=pwd, wait=True
                )

except ImportError:
    slixmpp = None


class XMPPCollectorBot(CollectorBot):
    name = "XMPP"
    xmpp = None
    collector_empty_process = True

    def init(self):
        self.logger.warning(
            "The output bot 'intelmq.bots.collectors.xmpp.output' "
            "is deprecated. It will be removed in version 3.0."
            "Please see https://github.com/certtools/intelmq/blob/"
            "develop/NEWS.md#xmpp-bots for more details."
        )
        if slixmpp is None:
            raise MissingDependencyError("slixmpp")

        # Retrieve Parameters from configuration
        xmpp_user = getattr(self.parameters, "xmpp_user", None)
        xmpp_server = getattr(self.parameters, "xmpp_server", None)
        xmpp_password = getattr(self.parameters, "xmpp_password", None)

        if None in (xmpp_user, xmpp_server, xmpp_password):
            raise ValueError("No User / Password provided.")
        else:
            xmpp_login = xmpp_user + "@" + xmpp_server

        self.userlist = getattr(self.parameters, "xmpp_userlist", [])
        # When configured in manager this is most likely a ,-separated string, we'd like an array
        if type(self.userlist) is str:
            self.userlist = [u.strip() for u in self.userlist.split(",")]
        elif self.userlist is None:  # if value is unset, set to empty list
            self.userlist = []

        self.whitelist_mode = getattr(self.parameters, "xmpp_whitelist_mode", False)

        self.muc = getattr(self.parameters, "use_muc", None)
        xmpp_room = getattr(self.parameters, "xmpp_room", None) if self.muc else None
        xmpp_room_nick = (
            getattr(self.parameters, "xmpp_room_nick", None) if self.muc else None
        )
        xmpp_room_password = (
            getattr(self.parameters, "xmpp_room_password", None) if self.muc else None
        )

        self.pass_full_xml = getattr(self.parameters, "pass_full_xml", None)
        self.strip_message = getattr(self.parameters, "strip_message", None)

        ca_certs = getattr(self.parameters, "ca_certs", None)

        if self.muc and not xmpp_room:
            raise ValueError("No room provided.")

        if self.muc:
            if not xmpp_room_nick:
                # create the room_nick from user and server
                xmpp_room_nick = xmpp_login

        self.xmpp = XMPPClient(
            xmpp_login,
            xmpp_password,
            xmpp_room,
            xmpp_room_nick,
            xmpp_room_password,
            self.logger,
        )

        if ca_certs:
            # Set CA-Certificates
            self.xmpp.ca_certs = ca_certs

        if self.xmpp.connect(reattempt=False):
            self.xmpp.process()
            # Add Handlers and register Plugins
            self.xmpp.register_plugin("xep_0030")  # Service Discovery
            self.xmpp.register_plugin("xep_0045")  # Multi-User Chat

            self.xmpp.add_event_handler("message", self.log_message)

        else:
            raise ValueError("Could not connect to XMPP-Server.")

    def process(self):
        # Processing is done by function called from the eventhandler...
        pass

    def shutdown(self):
        if self.xmpp:
            if self.xmpp.disconnect():
                self.logger.info("Disconnected from XMPP Server.")
            else:
                self.logger.error("Could not disconnect from XMPP Server.")
        else:
            self.logger.info("There was no XMPPClient I could stop.")

    def log_message(self, msg):
        # If some exception happens here, the bot would silently fail.
        # We want to know the reason, so we will log the exception manually.
        try:
            # Check if the message was sent by a users that is on
            # the white or blacklist, determine if the message shall
            # be processed.
            if self.muc:
                if msg["mucnick"] not in self.userlist and self.whitelist_mode:
                    # Whitelist-Case
                    return
                elif msg["mucnick"] in self.userlist and not self.whitelist_mode:
                    # Blacklist Case
                    return

            if self.pass_full_xml:
                body = str(msg)
            else:
                if self.strip_message:
                    body = msg["body"].strip()
                else:
                    body = msg["body"]

            if len(body) > 400:
                tmp_body = body[:397] + "..."
            else:
                tmp_body = body

            self.logger.debug("Received Stanza: %r from %r.", tmp_body, msg["from"])

            raw_msg = body
            # Read msg-body and add as raw to a new report.
            # now it's up to a parser to do the interpretation of the message.
            if raw_msg:
                report = self.new_report()
                report.add("raw", raw_msg)
                self.send_message(report)
        except Exception:
            self.logger.exception("Error during message handling.")
            raise


BOT = XMPPCollectorBot



"""Testing DISP Collector

SPDX-FileCopyrightText: 2023 CERT.at GmbH <https://cert.at/>
SPDX-License-Identifier: AGPL-3.0-or-later
"""

from datetime import datetime, timezone
from unittest import TestCase

import requests

from intelmq_extensions.bots.collectors.disp._client import DISPClient

from .base import DISP_APIMockMixIn


class DISPClientTestCase(DISP_APIMockMixIn, TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.config = {
            "api_url": "https://disp.example.at/v1",
            "auth_token": "XXX",
            "oauth_clientid": "client1",
            "oauth_clientsecret": "secret1",
            "oauth_url": "https://oauthip.example.at/v1",
        }
        self.setup_api_mock(
            url=self.config["api_url"],
            auth_token=self.config["auth_token"],
            oauth_clientid=self.config["oauth_clientid"],
            oauth_clientsecret=self.config["oauth_clientsecret"],
            oauth_url=self.config["oauth_url"],
            session=requests.Session(),
        )

        self.client = DISPClient(session=self.session, **self.config)

    def test_get_simple_response(self):
        self.mock_request(
            "incident/",
            json={"id": "some-id"},
        )

        response = self.client.get("incident/")
        self.assertEqual({"id": "some-id"}, response)

    def test_get_paginated_response_1_page(self):
        self.mock_request(
            "incident/?page=0&size=10",
            json={
                "content": [{"id": "id-1"}, {"id": "id-2"}],
                "last": True,
            },
        )

        responses = list(self.client.get_paginated("incident/"))
        self.assertEqual([{"id": "id-1"}, {"id": "id-2"}], responses)

    def test_get_paginated_response_3_pages(self):
        self.mock_request(
            "incident/?page=0&size=10",
            json={
                "content": [{"id": "id-1"}, {"id": "id-2"}],
                "last": False,
            },
        )
        self.mock_request(
            "incident/?page=1&size=10",
            json={
                "content": [{"id": "id-3"}],
                "last": False,
            },
        )
        self.mock_request(
            "incident/?page=2&size=10",
            json={
                "content": [{"id": "id-4"}],
                "last": True,
            },
        )

        responses = list(self.client.get_paginated("incident/"))
        self.assertEqual(
            [{"id": "id-1"}, {"id": "id-2"}, {"id": "id-3"}, {"id": "id-4"}], responses
        )

    def test_get_incidents(self):
        after = datetime(2023, 1, 1, 19, 10, tzinfo=timezone.utc)

        self.mock_request(
            "incident/?query=validationDate%20%3E%201672600200000%20AND%20UNREAD&page=0&size=10",
            json={"content": [{"id": "id-1"}]},
        )
        self.mock_request(
            "incident/?query=validationDate%20%3E%201672600200000&page=0&size=10",
            json={"content": [{"id": "id-2"}]},
        )
        self.mock_request(
            "incident/?query=NOT%20SOMETHING&page=0&size=10",
            json={"content": [{"id": "id-3"}]},
        )

        self.assertEqual([{"id": "id-1"}], list(self.client.incidents(after, True)))
        self.assertEqual([{"id": "id-2"}], list(self.client.incidents(after, False)))
        self.assertEqual(
            [{"id": "id-3"}], list(self.client.incidents(query="NOT SOMETHING"))
        )

    def test_get_evidence(self):
        # For credential tracking, the expected file is in JSON format.
        # For other incidents (not covered now), the format may differ
        self.mock_request("incident/id-1/file/f-1", text='{"credentials": []}')

        result = self.client.download_evidence_json("id-1", "f-1")

        self.assertEqual({"credentials": []}, result)

    def test_mark_read(self):
        self.mock_request("incident/read?id=id-1&read=true", method="post")

        self.client.mark_incident_read("id-1")

        self.assertEqual(1, self.requests.call_count)

    def test_raise_on_connection_issue(self):
        self.mock_request("/", status_code=500)

        with self.assertRaises(RuntimeError):
            self.client.get("/")

        self.mock_request("/", method="post", status_code=500)

        with self.assertRaises(RuntimeError):
            self.client.post("/")



# -*- coding: utf-8 -*-
"""
Squelcher Expert marks events as new or old depending on a TTL(ASN, Net, IP).
"""

from __future__ import unicode_literals

from ipaddress import ip_address, ip_network

from intelmq.lib.bot import Bot
from intelmq.lib.message import Event
from intelmq.lib.utils import load_configuration

try:
    import psycopg2
except ImportError:
    psycopg2 = None
try:
    import netaddr
except ImportError:
    netaddr = None

"""
If the event in the DB is older than 2 days, then we also check if it has been sent out.
If this is not the case, we assume the event will be sent out, thus we squelch the new event.
"""
SELECT_QUERY = """
SELECT COUNT(*) FROM {table}
WHERE
"time.source" >= LOCALTIMESTAMP - INTERVAL '%(ttl)s SECONDS' AND
"classification.type" = %(type)s AND
"classification.identifier" = %(identifier)s AND
{source_filters}
notify IS TRUE AND
("time.source" >= LOCALTIMESTAMP - INTERVAL %(sending_interval)s OR
 (sent_at IS NOT NULL AND "time.source" < LOCALTIMESTAMP - INTERVAL %(sending_interval)s)
)
"""

# If the event is newer than sending interval, assume it will be sent soon or already has been,
# regardless of the report id. If the source time is older, ignore events without report id as
# they were most probably forgotten
OPEN_REPORT_QUERY = """
SELECT COUNT(*) FROM {table}
WHERE
"time.source" >= LOCALTIMESTAMP - INTERVAL '%(ttl)s SECONDS' AND
"classification.type" = %(type)s AND
"classification.identifier" = %(identifier)s AND
{source_filters}
notify IS TRUE AND
(
    "time.source" >= LOCALTIMESTAMP - INTERVAL %(sending_interval)s OR
    (rtir_report_id IS NOT NULL AND "time.source" < LOCALTIMESTAMP - INTERVAL %(sending_interval)s)
)
"""

QUERY_MAP = {"base": SELECT_QUERY, "open_report": OPEN_REPORT_QUERY}
JSON_FIELDS = ["extra."]


class SquelcherExpertBot(Bot):
    configuration_path: str = ""
    connect_timeout: int = 5
    database: str = ""
    user: str = ""
    password: str = ""
    host: str = ""
    port: str = ""
    sslmode: str = ""
    autocommit: bool = True
    table: str = "contacts"
    sending_time_interval: int = 1
    overwrite: bool = False
    query: str = "base"  # base, open_report
    source_fields: str = "source.ip"
    filter_ip_only: str = True
    use_ttl_field: str = "extra.ttl"

    _filters_mapping: dict = None

    def init(self):
        self.config = load_configuration(self.configuration_path)

        self.logger.debug("Connecting to PostgreSQL.")
        if psycopg2 is None:
            raise ValueError("Could not import psycopg2. Please install it.")
        if netaddr is None:
            raise ValueError("Could not import netaddr. Please install it.")

        try:
            self.con = psycopg2.connect(
                database=self.database,
                user=self.user,
                password=self.password,
                host=self.host,
                port=self.port,
                sslmode=self.sslmode,
                connect_timeout=self.connect_timeout,
            )
            self.cur = self.con.cursor()
            self.con.autocommit = self.autocommit

        except Exception:
            self.logger.exception("Failed to connect to database.")
            self.stop()
        self.logger.info("Connected to PostgreSQL.")

        self.query_tpl = self._build_query()
        self.convert_config()

    def _build_query(self):
        # TODO: Build it using sql-safe syntax
        # TODO: ensure source field is in the harmonization
        template = QUERY_MAP[self.query]
        self._filters_mapping = dict()
        for idx, field in enumerate(self.source_fields.split(",")):
            self._filters_mapping[field] = f"source_{idx}"

        conditions = []
        for field, filter_key in self._filters_mapping.items():
            json_fields = [f for f in JSON_FIELDS if field.startswith(f)]
            if json_fields:
                conditions.append(
                    (
                        f'"{json_fields[0][:-1]}" ->> \'{field.replace(json_fields[0], "")}\''
                        f" IS NOT DISTINCT FROM %({filter_key})s AND"
                    )
                )
            else:
                # IS NOT DISTINCT makes a good job, but doesn't play well with
                # unique indexes unless probably PostgreSQL 15 and UNIQUE NULLS NOT DISTINCT
                # index configuration
                conditions.append(
                    f"""CASE WHEN %({filter_key})s IS null THEN
                            "{field}" is null
                            ELSE "{field}" = %({filter_key})s
                        END AND
                    """
                )
        return template.format(table=self.table, source_filters="\n".join(conditions))

    def convert_config(self):
        for rule_index, ruleset in enumerate(self.config):
            for key, value in ruleset[0].items():
                if isinstance(value, list):
                    self.config[rule_index][0][key] = tuple(value)
                if isinstance(value, dict):
                    self.config[rule_index][0][key] = tuple(value.items())

    def convert_event(self, event):
        event_copy = event.to_dict()
        for key, value in event_copy.items():
            if isinstance(value, list):
                event_copy[key] = tuple(value)
            if isinstance(value, dict):
                event_copy[key] = tuple(value.items())
        return event_copy

    def process(self):
        event = self.receive_message()

        if "notify" in event and not self.overwrite:
            self.logger.debug(
                "Notify field present and not allowed to overwrite, skipping."
            )
            self.modify_end(event)
            return

        if self.filter_ip_only:
            if "source.ip" not in event and "source.fqdn" in event:
                self.logger.debug(
                    "Filtering restricted to IPs, setting notify=true for domain event."
                )
                event.add("notify", True, overwrite=True)
                self.modify_end(event)
                return
            if "source.asn" not in event:
                self.logger.debug("Discarding event as it lacks AS number.")
                event.add("notify", False, overwrite=True)
                self.modify_end(event)
                return
        ttl = None

        if self.use_ttl_field:
            try:
                ttl = int(event[self.use_ttl_field])
            except KeyError:
                pass

        if ttl is None:
            for ruleset in self.config:
                condition = ruleset[0].copy()
                conditions = []
                if "source.network" in condition and "source.ip" in event:
                    conditions.append(
                        ip_address(event["source.ip"])
                        in ip_network(condition["source.network"])
                    )
                    del condition["source.network"]
                if "source.iprange" in condition and "source.ip" in event:
                    conditions.append(
                        event["source.ip"]
                        in netaddr.IPRange(*condition["source.iprange"])
                    )
                    del condition["source.iprange"]
                if set(condition.items()).issubset(
                    self.convert_event(event).items()
                ) and all(conditions):
                    ttl = ruleset[1]["ttl"]
                    break

        self.logger.debug(
            "Found TTL {} for ({}, {})." "".format(
                ttl, event.get("source.asn"), event.get("source.ip")
            )
        )

        try:
            if ttl >= 0:
                source_filters = {
                    filter_key: event.get(field)
                    for field, filter_key in self._filters_mapping.items()
                }
                self.cur.execute(
                    self.query_tpl,
                    {
                        "ttl": ttl,
                        "type": event["classification.type"],
                        "identifier": event["classification.identifier"],
                        "sending_interval": self.sending_time_interval,
                        **source_filters,
                    },
                )
                result = self.cur.fetchone()[0]
            else:  # never notify with ttl -1
                result = 1
        except (
            psycopg2.InterfaceError,
            psycopg2.InternalError,
            psycopg2.OperationalError,
            AttributeError,
        ):
            self.logger.exception("Cursor has been closed, connecting again.")
            self.init()
        else:
            if result == 0:
                notify = True
            else:
                notify = False

            event.add("notify", notify, overwrite=True)
            self.modify_end(event)

    def shutdown(self):
        try:
            self.cur.close()
        except Exception:
            pass
        try:
            self.con.close()
        except Exception:
            pass

    def modify_end(self, event):
        self.send_message(event)
        self.acknowledge_message()

    @staticmethod
    def check(parameters):
        retval = []
        try:
            config = load_configuration(parameters["configuration_path"])
        except ValueError as exc:
            return [["error", "Could not load configuration: %r." % exc]]
        for ruleset in config:
            condition = ruleset[0].copy()
            if "source.network" in condition:
                try:
                    ip_network(condition["source.network"])
                except ValueError as exc:
                    retval += [
                        [
                            "warning",
                            "%r is not a valid IP network: %r."
                            % (condition["source.network"], exc),
                        ]
                    ]
                del condition["source.network"]
            if "source.iprange" in condition:
                try:
                    netaddr.IPRange(*condition["source.iprange"])
                except ValueError as exc:
                    retval += [
                        [
                            "warning",
                            "%r is not a valid IP range: %r."
                            % (condition["source.iprange"], exc),
                        ]
                    ]
                del condition["source.iprange"]
            try:
                Event(condition)
            except Exception as exc:
                retval += [
                    ["warning", "Failed to parse conditions as Event: %r." % (exc)]
                ]
            try:
                int(ruleset[1]["ttl"])
            except ValueError as exc:
                retval += [
                    ["error", "%r is not a valid TTL: %r." % (ruleset[1]["ttl"], exc)]
                ]
        return retval if retval else None


BOT = SquelcherExpertBot



from intelmq.lib.test import skip_database

from intelmq_extensions.cli.create_reports import IntelMQCLIContoller

from ..base import CLITestCase


@skip_database()
class TestCreatingReports(CLITestCase):
    CLI_CONTROLLER = IntelMQCLIContoller

    def setUp(self) -> None:
        super().setUp()
        self.event_ids = [
            self.db_add_event({"feed.code": "feed1", "constituency": "energy"}),
            self.db_add_event({"feed.code": "feed2", "constituency": "energy"}),
            self.db_add_event({"feed.code": "feed3", "constituency": "national"}),
            self.db_add_event({"feed.code": "feed4", "constituency": "national"}),
            self.db_add_event({"feed.code": "feed5", "constituency": None}),
        ]

    def test_list_open_feeds(self):
        self.config.update({"constituency": {"key": "energy"}})

        self.run_cli(["-l"])

        self.assertEqual(2, len(self.stdout))
        self.assertIn("feed1\n", self.stdout)
        self.assertIn("feed2\n", self.stdout)

        self.config.update({"constituency": {"key": "national", "default": True}})

        self.run_cli(["-l"])
        self.assertEqual(3, len(self.stdout))
        self.assertIn("feed3\n", self.stdout)
        self.assertIn("feed4\n", self.stdout)
        self.assertIn("feed5\n", self.stdout)

    def test_run_create_reports(self):
        self.config.update({"constituency": {"key": "energy"}})

        self.run_cli([])

        self.assertInLogs("Handling feedcode 'feed1'")
        self.assertInLogs("Handling feedcode 'feed2'")
        self.assertRTTicketCount(2)

        event_feed1 = self.db_get_event(self.event_ids[0])
        self.assertEqual(event_feed1["rtir_report_id"], 1)

        event_feed2 = self.db_get_event(self.event_ids[1])
        self.assertEqual(event_feed2["rtir_report_id"], 2)

        self.assertIn("Owner", self.rt_mock.create_ticket.call_args.kwargs)

    def test_create_internal_only_notifications(self):
        self.config.update({"constituency": {"key": "national", "default": True}})
        id_1 = self.db_add_event(
            {"feed.code": "internal-only"},
            notify=True,
            abuse_contact="",
            cc="PL",
            extra={"monitored_asset": "something"},
        )
        id_2 = self.db_add_event(
            {"feed.code": "internal-only2"}, notify=True, abuse_contact="", cc="AT"
        )
        id_3 = self.db_add_event(
            {"feed.code": "internal-only2"},
            notify=True,
            abuse_contact="",
            cc="US",
            extra={"monitored_asset": "something"},
        )

        self.run_cli(
            [
                "--internal-notify",
                "--monitored-assets",
                "--feed",
                "internal-only",
                "internal-only2",
            ]
        )

        self.assertRTTicketCount(2)
        self.assertEqual(self.db_get_event(id_1)["rtir_report_id"], 1)
        for event_id in [id_2, id_3]:
            event = self.db_get_event(event_id)
            self.assertEqual(event["rtir_report_id"], 2)

        self.assertNotIn("Owner", self.rt_mock.create_ticket.call_args.kwargs)

        # Other events are untouched
        for event_id in self.event_ids:
            event = self.db_get_event(event_id)
            self.assertIsNone(event["rtir_report_id"])



"""Transformers for the BlackKite

SPDX-FileCopyrightText: 2023 CERT.at GmbH <https://cert.at/>
SPDX-License-Identifier: AGPL-3.0-or-later
"""

from abc import ABC
from datetime import datetime
from typing import Iterable, Sequence

import dateutil.parser
from dateutil.tz import UTC
from intelmq.lib.harmonization import IPAddress

from intelmq_extensions.lib.blackkite import Category


class BaseTransformer(ABC):
    DEFAULT_CLASSIFICATION = ("", "", "")
    CLASSIFICATION_MAP: dict[str, tuple] = {}
    _COMMON_FIELDS = (
        ("extra.feed_event_id", "FindingId"),
        # In general, the Domain field does not represent the affected domain.
        # It can be a primary domain of the affected URL, or just the domain
        # configured for the given company, without direct relation to the finding.
        # FIXME: waiting for BlackKite response on how to get the affected URL
        ("source.fqdn", "", "get_domain"),
        ("source.ip", "IpAddress", "get_ip"),
        ("event_description.text", "Detail", "get_description"),
        # The date that Black Kite first seen the finding.
        ("time.source", "FindingDate"),
        ("feed.documentation", "ControlId", "get_documentation_url"),
    )
    SPECIFIC_FIELDS = ()

    def __init__(self) -> None:
        self._data = None
        self._transformed = {}

    def to_events_data(self, incident_data: dict) -> Iterable[dict]:
        self._data = incident_data
        self._transformed = {}
        self._finding_id = incident_data.get("ControlId")
        self.transform()
        for event_data in self.transform_single():
            yield event_data

    def _map(self, event_key: str, finding_key: str, mapper: str = None):
        if finding_key in self._data:
            if not mapper:
                self._transformed[event_key] = self._data[finding_key]
            else:
                self._transformed[event_key] = getattr(self, mapper)(finding_key)

    def _map_many(self, mappings: Sequence[tuple]):
        for mapping in mappings:
            self._map(*mapping)

    def _map_classification(self):
        taxonomy, type_, identifier = self.CLASSIFICATION_MAP.get(
            self._finding_id, self.DEFAULT_CLASSIFICATION
        )
        self._transformed["classification.taxonomy"] = taxonomy
        self._transformed["classification.type"] = type_
        self._transformed["classification.identifier"] = identifier

    def _map_feed_data(self):
        category = self._data.get("ControlId", "").split("-")[0]
        self._transformed["feed.code"] = f"blackkite-{category.lower()}"
        self._transformed["feed.name"] = f"BlackKite {category.upper()}"

    def get_description(self, _):
        description = []
        if title := self._data.get("Title"):
            description.append(title)
        if detail := self._data.get("Detail"):
            description.append(detail)
        return ". ".join(description)

    def transform(self):
        self._map_classification()
        self._map_feed_data()
        self._map_many(self._COMMON_FIELDS)
        self._map_many(self.SPECIFIC_FIELDS)

    def transform_single(self) -> Iterable[dict]:
        """If an incident produce multiple events,
        this method should do the extraction and transformation"""

        yield self._transformed

    def get_documentation_url(self, _):
        return f"https://cyber.riskscore.cards/kb/{self._finding_id}"

    def lower(self, field: str) -> str:
        return (self._data.get(field) or "").lower()

    def get_ip(self, field: str):
        # BlackKite can send domain in the IPAddress field
        ip_data = self._data.get(field)
        if IPAddress.is_valid(ip_data, sanitize=True):
            return ip_data
        return None

    def get_domain(self, field: str):
        # BlackKite can send domain in the IPAddress field
        # as it's then a better domain, this should be used
        if ip_data := self._data.get("IpAddress"):
            if not IPAddress.is_valid(ip_data):
                return ip_data

        if field:
            return self._data.get(field)

    def get_first(self, field: str) -> str:
        """Get first item from the list"""
        items = self._data.get(field)
        if items:
            return items[0]
        return None


class PatchManagementTransformer(BaseTransformer):
    DEFAULT_CLASSIFICATION = ("vulnerable", "vulnerable-system", "bk-patchmanagement")
    CLASSIFICATION_MAP = {
        "PATCH-010": ("vulnerable", "vulnerable-system", "end-of-live")
    }
    SPECIFIC_FIELDS = (
        ("extra.product_name", "ProductName"),
        # lower to match with data from other sources
        ("extra.vulnerabilities", "CveId", "lower"),
        ("event_description.url", "References", "get_first"),
    )

    def _map_cpe(self):
        cpes = self._data.get("Cpes")
        if not cpes:
            return
        cpe_data = cpes[0].split(":")
        self._transformed["extra.vendor"] = cpe_data[3]
        self._transformed["extra.product"] = cpe_data[4]

    def _map_cve_to_identifier(self):
        if cve_id := self._data.get("CveId"):
            self._transformed["classification.identifier"] = cve_id.lower()

    def transform(self) -> Iterable[dict]:
        super().transform()
        self._map_cpe()
        self._map_cve_to_identifier()


class ApplicationSecurityTransformer(BaseTransformer):
    CLASSIFICATION_MAP = {
        "APPSEC-014": ("vulnerable", "potentially-unwanted-accessible", "bk-appsec")
    }


class CredentialManagementTransformer(BaseTransformer):
    DEFAULT_CLASSIFICATION = (
        "information-content-security",
        "data-leak",
        "leaked-credentials",
    )
    SPECIFIC_FIELDS = (
        ("source.fqdn", "EmailorUsername", "get_domain_from_email"),
        ("source.account", "EmailorUsername"),
        ("extra.account", "EmailorUsername"),
        ("extra.password", "PasswordType"),
        ("extra.compromise_time_full", "LeakDate"),
        ("extra.compromise_time", "LeakDate", "get_compromise_time"),
        ("extra.leak_source", "Source"),
    )
    _DESCRIPTION = (
        "A user with email in your domain"
        " was found in leaked credentials related to: {source}"
    )
    _COMPROMISE_TIME_FORMAT = "%Y-%m"

    def get_description(self, _):
        return self._DESCRIPTION.format(source=self._data.get("Source"))

    def get_compromise_time(self, _):
        # return the month only, as in DISP parser
        date_str = self._data.get("LeakDate")
        if not date_str:
            return None
        date: datetime = dateutil.parser.parse(date_str).astimezone(tz=UTC)
        return date.strftime(self._COMPROMISE_TIME_FORMAT)

    def get_domain_from_email(self, field):
        data = self._data.get(field)
        if "@" in data:
            return data.split("@")[-1]
        return None


TRANSFORMERS_MAPPER: dict[Category, BaseTransformer] = {
    Category.PatchManagement: PatchManagementTransformer(),
    Category.ApplicationSecurity: ApplicationSecurityTransformer(),
    Category.CredentialManagement: CredentialManagementTransformer(),
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" """

import csv
import datetime
import io
import locale
import os
import readline  # noqa: F401 hooks into input()
import subprocess
import sys
import tempfile
import zipfile
from functools import lru_cache
from time import sleep

import tabulate
from termstyle import bold, inverted, reset

from intelmq_extensions.cli import lib

myinverted = str(reset) + str(inverted)


class IntelMQCLIContoller(lib.IntelMQCLIContollerTemplate):
    appname = "intelmqcli"
    usage = lib.USAGE
    epilog = lib.EPILOG
    table_mode = False  # for sticky table mode
    dryrun = False
    verbose = False
    batch = False
    compress_csv = False
    boilerplate = None
    zipme = False
    subject = None

    def __init__(self, overridden_config: dict = None):
        super().__init__(overridden_config)
        self._collected_descriptions = set()

    def run(self, args: list):
        self.parser.add_argument(
            "-l", "--list-feeds", action="store_true", help="List all feeds"
        )
        self.parser.add_argument(
            "-i", "--list-identifiers", action="store_true", help="List all identifiers"
        )
        self.parser.add_argument(
            "-L", "--list-texts", action="store_true", help="List all existing texts."
        )
        self.parser.add_argument(
            "-t", "--text", nargs=1, help="Specify the text to be used."
        )
        self.parser.add_argument(
            "-s",
            "--subject",
            nargs=1,
            help="Specify the subject to be used instead of the per-taxonomy.",
        )
        self.parser.add_argument(
            "-T", "--list-taxonomies", action="store_true", help="List all taxonomies"
        )
        self.parser.add_argument(
            "-y", "--list-types", action="store_true", help="List all types"
        )

        self.parser.add_argument(
            "-c",
            "--compress-csv",
            action="store_true",
            help="Automatically compress/shrink the attached CSV report if"
            " fields are empty (default = False).",
        )
        self.parser.add_argument(
            "-z",
            "--zip",
            action="store_true",
            help="Zip every events.csv attachement to an "
            "investigation for RT (defaults to false)",
        )
        self.setup(args)

        if self.args.compress_csv:
            self.compress_csv = True
        if self.args.text:
            self.boilerplate = self.args.text[0]
        if self.args.zip:
            self.zipme = True
        if self.args.subject:
            self.subject = self.args.subject[0]

        self.connect_database()

        if self.args.list_feeds:
            self.execute(lib.QUERY_FEED_NAMES, extend=False)
            for row in self.cur.fetchall():
                if row["feed.code"]:
                    print(row["feed.code"])
            return 0

        if self.args.list_texts:
            self.execute(lib.QUERY_TEXT_NAMES, extend=False)
            for row in self.cur.fetchall():
                if row["key"]:
                    print(row["key"])
            return 0

        if self.args.list_identifiers:
            self.execute(lib.QUERY_IDENTIFIER_NAMES, extend=False)
            for row in self.cur.fetchall():
                if row["classification.identifier"]:
                    print(row["classification.identifier"])
            return 0

        if self.args.list_taxonomies:
            self.execute(lib.QUERY_TAXONOMY_NAMES, extend=False)
            for row in self.cur.fetchall():
                if row["classification.taxonomy"]:
                    print(row["classification.taxonomy"])
            return 0

        if self.args.list_types:
            self.execute(lib.QUERY_TYPE_NAMES, extend=False)
            for row in self.cur.fetchall():
                if row["classification.type"]:
                    print(row["classification.type"])
            return 0

        if locale.getpreferredencoding() != "UTF-8":
            self.logger.error(
                "The preferred encoding of your locale setting is not UTF-8 "
                "but %s. Exiting.",
                locale.getpreferredencoding(),
            )
            return 1

        if not self.rt.login():
            self.logger.error(
                "Could not login as %r on %r.",
                self.config["rt"]["user"],
                self.config["rt"]["uri"],
            )
            return 2
        else:
            self.logger.info(
                "Logged in as %r on %r.",
                self.config["rt"]["user"],
                self.config["rt"]["uri"],
            )
        try:
            self.execute(lib.QUERY_OPEN_TAXONOMIES)
            taxonomies = [x["classification.taxonomy"] for x in self.cur.fetchall()]
            self.logger.info("All taxonomies: " + ", ".join(taxonomies))
            for taxonomy in taxonomies:
                self.logger.info("Handling taxonomy %r.", taxonomy)
                if (
                    taxonomy not in lib.SUBJECT or lib.SUBJECT[taxonomy] is None
                ) and not self.subject:
                    self.logger.error("No subject defined for %r." % taxonomy)
                    continue
                self.execute(lib.QUERY_OPEN_EVENT_REPORTS_BY_TAXONOMY, (taxonomy,))
                report_ids = [x["rtir_report_id"] for x in self.cur.fetchall()]
                self.execute(lib.QUERY_OPEN_EVENT_IDS_BY_TAXONOMY, (taxonomy,))
                event_ids = [x["id"] for x in self.cur.fetchall()]
                if self.subject:
                    subject = self.subject
                else:
                    subject = "%s %s incidents on %s" "" % (
                        len(event_ids),
                        lib.SUBJECT[taxonomy],
                        datetime.datetime.now().strftime("%Y-%m-%d"),
                    )

                if self.dryrun:
                    self.logger.info("Simulate creation of incident.")
                    incident_id = -1
                else:
                    incident_id = self.rt.create_ticket(
                        Queue="Incidents",
                        Subject=subject,
                        Owner=self.config["rt"]["user"],
                    )
                    if incident_id == -1:
                        self.logger.error("Could not create Incident %r.", subject)
                        continue

                    self.logger.info("Created Incident %s." % incident_id)
                    # XXX TODO: distinguish between national and other constituencies
                    self.rt.edit_ticket(
                        incident_id,
                        CF_Classification=taxonomy,
                        # CF_Constituency='NATIONAL',
                        CF_Function="IncidentCoord",
                    )

                for report_id in report_ids:
                    if not self.dryrun and not self.rt.edit_link(
                        report_id, "MemberOf", incident_id
                    ):
                        self.logger.error(
                            "Could not link Incident to Incident Report: (%d -> %d).",
                            incident_id,
                            report_id,
                        )
                        continue
                    elif self.dryrun:
                        self.logger.info(
                            "Would have linked Incident Report %d to Incident.",
                            report_id,
                        )

                self.executemany(
                    "UPDATE {events} SET rtir_incident_id = %s WHERE id = %s",
                    [(incident_id, event_id) for event_id in event_ids],
                    extend=False,
                )
                self.con.commit()
                self.logger.info("Linked events to incident.")

                if not self.dryrun:
                    self.execute(
                        lib.QUERY_DISTINCT_CONTACTS_BY_INCIDENT, (incident_id,)
                    )
                else:
                    self.execute(
                        lib.DRY_QUERY_DISTINCT_CONTACTS_BY_TAXONOMY, (taxonomy,)
                    )

                contacts = [x["source.abuse_contact"] for x in self.cur.fetchall()]
                inv_results = []

                for contact in contacts:
                    self.logger.info("Handling contact " + contact)
                    if not self.dryrun:
                        self.execute(
                            lib.QUERY_EVENTS_BY_ASCONTACT_INCIDENT,
                            (
                                incident_id,
                                contact,
                            ),
                        )
                    else:
                        self.execute(
                            lib.DRY_QUERY_EVENTS_BY_ASCONTACT_TAXONOMY,
                            (
                                taxonomy,
                                contact,
                            ),
                        )
                    data = self.cur.fetchall()
                    results = self.send(taxonomy, contact, data, incident_id)
                    if results:
                        inv_results.append(results)
                    else:
                        return 1

                if all(inv_results):
                    try:
                        if not self.dryrun and not self.rt.edit_ticket(
                            incident_id, Status="resolved"
                        ):
                            self.logger.error(
                                "Could not close incident %d.", incident_id
                            )
                    except IndexError:
                        # Bug in RT/python-rt
                        pass
                else:
                    self.logger.warn(
                        "Not all investigations completed -> Can't resolve "
                        "incident %d.",
                        incident_id,
                    )
            self.execute(lib.QUERY_HALF_PROC_INCIDENTS)
            query = [
                (x["rtir_incident_id"], x["classification.taxonomy"])
                for x in self.cur.fetchall()
            ]
            self.logger.info("All half processed incidents and taxonomy: " + str(query))
            for incident_id, taxonomy in query:
                self.logger.info(
                    "Handling incident %d and taxonomy %r.", incident_id, taxonomy
                )
                if (
                    taxonomy not in lib.SUBJECT or lib.SUBJECT[taxonomy] is None
                ) and not self.args.subject:
                    self.logger.error("No subject defined for %r." % taxonomy)
                    continue

                self.execute(lib.QUERY_DISTINCT_CONTACTS_BY_INCIDENT, (incident_id,))
                contacts = [x["source.abuse_contact"] for x in self.cur.fetchall()]

                inv_results = []
                for contact in contacts:
                    self.logger.info("Handling contact " + contact)
                    self.execute(
                        lib.QUERY_EVENTS_BY_ASCONTACT_INCIDENT,
                        (
                            incident_id,
                            contact,
                        ),
                    )
                    data = self.cur.fetchall()
                    results = self.send(taxonomy, contact, data, incident_id)
                    if results:
                        inv_results.append(results)
                    else:
                        return 1

                if all(inv_results):
                    # This is a terrible solution, but looks like RT has a race-condition causing
                    # re-opening incidents if we are too quick.
                    sleep(5)
                    try:
                        if not self.dryrun and not self.rt.edit_ticket(
                            incident_id, Status="resolved"
                        ):
                            self.logger.error(
                                "Could not close incident %d.", incident_id
                            )
                    except IndexError:
                        # Bug in RT/python-rt
                        pass
                else:
                    self.logger.warn(
                        "Not all investigations completed -> Can't resolve "
                        "incident %d.",
                        incident_id,
                    )
                    return 1
            return 0

        finally:
            self.rt.logout()

    def query_get_text(self, text_id):
        self.execute(
            lib.QUERY_GET_TEXT.format(texttab=self.config["database"]["text_table"]),
            (text_id,),
            extend=False,
        )

    def get_text(self, text_id):
        text = None
        if self.boilerplate:  # get id from parameter
            text_id = self.boilerplate
        self.query_get_text(text_id)
        if self.cur.rowcount:
            text = self.cur.fetchall()[0]["body"]
        if not text:  # if all failed, get the default
            self.query_get_text(self.config["database"]["default_key"])
            if self.cur.rowcount:
                text = self.cur.fetchall()[0]["body"]
            else:
                self.logger.error("Default text not found!")
                return None

        return text

    def shrink_dict(self, d):
        if not self.compress_csv:
            return d
        keys = list(d[0].keys())
        empty = dict(zip(keys, [True] * len(keys)))
        for line in d:
            for key, value in line.items():
                if value is not None:
                    empty[key] = False
        return [{k: v for k, v in dicti.items() if not empty[k]} for dicti in d]

    def _preprocess_row(self, row):
        description = row.get("event_description.text")
        if description:
            self._collected_descriptions.add(description.replace("\\n", "\n"))
            # \\n - encoded new lines
            row["event_description.text"] = description.replace("\n", " ").replace(
                "\\n", " "
            )
        return row

    @property
    @lru_cache()
    def event_description_divider(self):
        divider = None
        self.query_get_text("event-description-divider")
        if self.cur.rowcount:
            divider = self.cur.fetchall()[0]["body"]
        return divider or "\n --- \n"

    def _prepare_descriptions_text(self):
        return self.event_description_divider.join(self._collected_descriptions)

    def send(self, taxonomy, contact, query, incident_id, requestor=None):
        if not query:
            self.logger.error("No data!")
            return False
        if not requestor:
            requestor = contact

        # PREPARATION
        query = self.shrink_dict(query)
        ids = list(str(row["id"]) for row in query)

        if self.subject:
            subject = self.subject
        else:
            subject = "{tax} incidents in your network: {date}" "".format(
                date=datetime.datetime.now().strftime("%Y-%m-%d"),
                tax=lib.SUBJECT[taxonomy],
            )
        text = self.get_text(taxonomy) or ""
        csvfile = io.StringIO()
        writer = csv.DictWriter(
            csvfile,
            fieldnames=lib.CSV_FIELDS,
            quoting=csv.QUOTE_MINIMAL,
            delimiter=str(";"),
            extrasaction="ignore",
            lineterminator="\n",
        )
        writer.writeheader()
        query_unicode = query
        self._collected_descriptions = set()
        writer.writerows(self._preprocess_row(row) for row in query)
        event_descriptions = self._prepare_descriptions_text()
        text = text.format(event_descriptions=event_descriptions)

        # note this might contain UTF-8 chars! let's ignore utf-8 errors. sorry.
        attachment_text = csvfile.getvalue()
        attachment_lines = attachment_text.splitlines()

        if self.verbose:
            self.logger.info(text)

        showed_text = (
            "=" * 100
            + """
To: {to}
Subject: {subj}

{text}
    """.format(
                to=requestor, subj=subject, text=text
            )
        )
        showed_text_len = showed_text.count("\n")

        # SHOW DATA
        if self.table_mode:
            if self.quiet:
                height = 80  # assume anything for quiet mode
            else:
                height = lib.getTerminalHeight() - 3 - showed_text_len
            csvfile.seek(0)
            if len(query) > height:
                with tempfile.NamedTemporaryFile(mode="w+") as handle:
                    handle.write(showed_text + "\n")
                    handle.write(
                        tabulate.tabulate(query, headers="keys", tablefmt="psql")
                    )
                    handle.seek(0)
                    subprocess.call(["less", handle.name])
            else:
                self.logger.info(showed_text)
                self.logger.info(
                    tabulate.tabulate(query_unicode, headers="keys", tablefmt="psql")
                )
        else:
            if self.quiet:
                height = 80
            else:
                height = lib.getTerminalHeight() - 4
            if 5 + len(query) > height:  # cut query too, 5 is length of text
                self.logger.info("\n".join(showed_text.splitlines()[:5]))
                self.logger.info("...")
                self.logger.info("\n".join(attachment_lines[: height - 5]))
                self.logger.info("...")
            elif showed_text_len + len(query) > height > 5 + len(query):
                self.logger.info(
                    "\n".join(showed_text.splitlines()[: height - len(query)])
                )
                self.logger.info("...")
                self.logger.info(attachment_text)
            else:
                self.logger.info(showed_text)
                self.logger.info(attachment_text)
        self.logger.info("-" * 100)

        # MENU
        if self.batch and requestor:
            answer = "s"
        else:
            answer = "q"
            if self.batch:
                self.logger.error("You need to set a valid requestor!")
            else:
                answer = input(
                    "{i}{b}[a]{i}utomatic, {b}[n]{i}ext, {i}{b}[s]{i}end, show "
                    "{b}[t]{i}able, change {b}[r]{i}equestor or {b}[q]{i}uit?{r} "
                    "".format(b=bold, i=myinverted, r=reset)
                ).strip()
        if answer == "q":
            exit(0)
        elif answer == "n":
            return False
        elif answer == "a":
            self.batch = True
        elif answer == "t":
            self.table_mode = bool((self.table_mode + 1) % 2)
            return self.send(taxonomy, contact, query, incident_id, requestor)
        elif answer == "r":
            answer = input(inverted("New requestor address:") + " ").strip()
            if len(answer) == 0:
                requestor = contact
            else:
                requestor = answer
            return self.send(taxonomy, contact, query, incident_id, requestor)
        elif answer != "s":
            self.logger.error("Unknow command %r.", answer)
            return self.send(taxonomy, contact, query, incident_id, requestor)

        if text is None:
            self.logger.error("I won't send with a missing text!")
            return False

        # INVESTIGATION
        if self.dryrun:
            self.logger.info("Simulate creation of investigation.")
            investigation_id = -1
        else:
            investigation_id = self.rt.create_ticket(
                Queue="Investigations",
                Subject=subject,
                Owner=self.config["rt"].get(
                    "investigation_owner", self.config["rt"]["user"]
                ),
                Requestor=requestor,
            )

            if investigation_id == -1:
                self.logger.error("Could not create Investigation %r.", subject)
                return False

            self.logger.info("Created Investigation %d.", investigation_id)
            if not self.rt.edit_link(incident_id, "HasMember", investigation_id):
                self.logger.error("Could not link Investigation to Incident.")
                return False

            self.executemany(
                "UPDATE {events} SET rtir_investigation_id = %s WHERE id = %s",
                [(investigation_id, evid) for evid in ids],
                extend=False,
            )
            self.logger.info("Linked events to investigation.")

        # CORRESPOND
        filename = "%s-%s.csv" % (
            datetime.datetime.now().strftime("%Y-%m-%d"),
            taxonomy,
        )
        if self.zipme or len(query) > self.config["rt"]["zip_threshold"]:
            attachment = io.BytesIO()
            ziphandle = zipfile.ZipFile(
                attachment, mode="w", compression=zipfile.ZIP_DEFLATED
            )
            data = csvfile.getvalue()
            ziphandle.writestr("events.csv", data)
            ziphandle.close()
            attachment.seek(0)
            filename += ".zip"
            mimetype = "application/octet-stream"
        else:
            attachment = csvfile
            attachment.seek(0)
            mimetype = "text/csv"

        try:
            # TODO: CC
            if self.dryrun:
                self.logger.info("Simulate creation of correspondence.")
            else:
                correspond = self.rt.reply(
                    investigation_id,
                    text=text,
                    files=[(filename, attachment, mimetype)],
                )
                if not correspond:
                    self.logger.error("Could not correspond with text and file.")
                    return False
                self.logger.info("Correspondence added to Investigation.")

            self.execute(
                "UPDATE {events} SET sent_at = LOCALTIMESTAMP WHERE "
                "rtir_investigation_id = %s",
                (investigation_id,),
                extend=False,
            )
            self.logger.info("Marked events as sent.")
        except Exception:
            self.con.rollback()
            raise
        else:
            self.con.commit()

            # RESOLVE
            try:
                if not self.dryrun and not self.rt.edit_ticket(
                    investigation_id, Status="resolved"
                ):
                    self.logger.error(
                        "Could not close investigation %d.", investigation_id
                    )
            except IndexError:
                # Bug in RT/python-rt
                pass

        if requestor != contact:
            asns = set(str(row["source.asn"]) for row in query)
            answer = input(
                inverted(
                    "Save recipient {!r} for ASNs {!s}? [Y/n] "
                    "".format(requestor, ", ".join(asns))
                )
            ).strip()
            if answer.strip().lower() in ("", "y", "j"):
                self.executemany(
                    lib.QUERY_UPDATE_CONTACT,
                    [(requestor, asn) for asn in asns],
                    extend=False,
                )
                self.con.commit()
                if self.cur.rowcount == 0:
                    self.query_insert_contact(asns=asns, contact=requestor)

        return True

    def query_insert_contact(self, contact, asns):
        user = os.environ["USER"]
        time = datetime.datetime.now().strftime("%c")
        comment = "Added by {user} @ {time}".format(user=user, time=time)
        self.executemany(
            lib.QUERY_INSERT_CONTACT,
            [(asn, contact, comment) for asn in asns],
            extend=False,
        )
        self.con.commit()


def main():
    controller = IntelMQCLIContoller()
    sys.exit(controller.run(sys.argv[1:]))


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
Testing certat_contact
"""
import os
import unittest

import intelmq.lib.test as test

from intelmq_extensions.bots.experts.certat_contact_intern.expert import (
    CERTatContactExpertBot,
)

from ....base import POSTGRES_CONFIG, BotTestCase

if os.environ.get("INTELMQ_TEST_DATABASES"):
    import psycopg2
    from psycopg2 import sql

INPUT1 = {
    "__type": "Event",
    "source.asn": 64496,
    "time.observation": "2015-01-01T00:00:00+00:00",
    "feed.code": "another-feed-code",
}
OUTPUT1 = {
    "__type": "Event",
    "source.asn": 64496,
    "source.abuse_contact": "cert@example.com",
    "feed.code": "another-feed-code",
    "time.observation": "2015-01-01T00:00:00+00:00",
    "destination_visible": True,
}
INPUT2 = {
    "__type": "Event",
    "source.asn": 64496,
    "time.observation": "2015-01-01T00:00:00+00:00",
    "feed.code": "example-feed",
}
OUTPUT2 = {
    "__type": "Event",
    "source.asn": 64496,
    "source.abuse_contact": "cert@example.com",
    "time.observation": "2015-01-01T00:00:00+00:00",
    "feed.code": "example-feed",
    "destination_visible": False,
}
OUTPUT3 = {
    "__type": "Event",
    "source.asn": 64496,
    "source.abuse_contact": "cert@example.com",
    "time.observation": "2015-01-01T00:00:00+00:00",
    "feed.code": "example-feed",
    "destination_visible": True,
}
INPUT4 = {
    "__type": "Event",
    "source.asn": 64497,
    "time.observation": "2015-01-01T00:00:00+00:00",
    "feed.code": "another-feed-code",
}
OUTPUT4 = {
    "__type": "Event",
    "source.asn": 64497,
    "time.observation": "2015-01-01T00:00:00+00:00",
    "feed.code": "another-feed-code",
    "destination_visible": True,
}


@test.skip_database()
class TestCERTatContactExpertBot(BotTestCase, unittest.TestCase):
    """
    A TestCase for CERTatContactExpertBot.
    """

    @classmethod
    def set_bot(cls):
        cls.bot_reference = CERTatContactExpertBot
        cls.default_input_message = INPUT1
        if not os.environ.get("INTELMQ_TEST_DATABASES"):
            return
        cls.sysconfig = {
            "autocommit": True,
            "ascolumn": "asn",
            "column": "contact",
            "feed_code": "example-feed",
            "table": "test_contacts",
            "overwrite": False,
        }
        cls.sysconfig.update(POSTGRES_CONFIG)
        cls.con = psycopg2.connect(
            database=cls.sysconfig["database"],
            user=cls.sysconfig["user"],
            password=cls.sysconfig["password"],
            host=cls.sysconfig["host"],
            port=cls.sysconfig["port"],
            sslmode=cls.sysconfig["sslmode"],
        )
        cls.con.autocommit = True
        cls.cur = cls.con.cursor()
        cls.create_if_not_exists(cls)
        cls.truncate(cls)

    def create_if_not_exists(self):
        self.cur.execute(
            sql.SQL(
                "CREATE TABLE IF NOT EXISTS {} ({} integer, {} text, {} boolean);"
            ).format(
                sql.Identifier(self.sysconfig["table"]),
                sql.Identifier(self.sysconfig["ascolumn"]),
                sql.Identifier(self.sysconfig["column"]),
                sql.Identifier(
                    "can-see-tlp-amber_{}".format(self.sysconfig["feed_code"])
                ),
            ),
        )

    def truncate(self):
        self.cur.execute("TRUNCATE TABLE {}".format(self.sysconfig["table"]))

    def insert(self, asn, contact, tlp_amber):
        query = """
INSERT INTO {table}(
    "{ascolumn}", "{column}", "can-see-tlp-amber_{feed_code}"
) VALUES (%s, %s, %s)
""".format(
            table=self.sysconfig["table"],
            column=self.sysconfig["column"],
            feed_code=self.sysconfig["feed_code"],
            ascolumn=self.sysconfig["ascolumn"],
        )
        self.cur.execute(query, (asn, contact, tlp_amber))

    def test_simple(self):
        "simple query"
        self.insert(64496, "cert@example.com", False)
        self.input_message = INPUT1
        self.run_bot()
        self.assertMessageEqual(0, OUTPUT1)

    def test_special_feed(self):
        "query with special feed code"
        self.insert(64496, "cert@example.com", False)
        self.input_message = INPUT2
        self.run_bot()
        self.assertMessageEqual(0, OUTPUT2)

    def test_special_feed_amber(self):
        "query with special feed code"
        self.insert(64496, "cert@example.com", True)
        self.input_message = INPUT2
        self.run_bot()
        self.assertMessageEqual(0, OUTPUT3)

    def test_no_result_destination_visible(self):
        "check if destination_visible to set to true if there's no match in the DB"
        self.input_message = INPUT4
        self.run_bot()
        self.assertMessageEqual(0, OUTPUT4)

    def tearDown(self):
        self.truncate()
        super().tearDown()

    @classmethod
    def tearDownClass(cls):
        if not os.environ.get("INTELMQ_TEST_DATABASES"):
            return
        cls.truncate(cls)
        cls.cur.close()
        cls.con.close()


if __name__ == "__main__":  # pragma: no cover
    unittest.main()



"""Common helpers to speak with APIs

SPDX-FileCopyrightText: 2023 CERT.at GmbH <https://cert.at/>
SPDX-License-Identifier: AGPL-3.0-or-later
"""

import base64
import json
import logging
import time
from contextlib import contextmanager
from datetime import datetime, timedelta
from threading import RLock

import requests
from dateutil.tz import UTC

default_logger = logging.getLogger(__name__)
# refresh access token when it's about to expire in 10 minutes
DEFAULT_REFRESH_WINDOW = 10


class OAuthAccessMixin:
    def init_oauth(
        self,
        oauth_clientid: str,
        oauth_clientsecret: str,
        oauth_url: str,
        session: requests.Session,
        oauth_scope: str = "",
        oauth_grant_type: str = "client_credentials",
        logger: logging.Logger = default_logger,
        refresh_before: int = DEFAULT_REFRESH_WINDOW,
        limiter: "RateLimiter" = None,
    ) -> None:
        self._oauth_clientid = oauth_clientid
        self._oauth_clientsecret = oauth_clientsecret
        self._oauth_url = oauth_url
        self._oauth_scope = oauth_scope
        self._oauth_grant_type = oauth_grant_type
        self._refresh_before = refresh_before
        self._access_token = None
        self.__logger = logger
        self.__session = session
        self.__limiter = limiter or RateLimiter(0, 0)

    @property
    def access_token(self):
        if self._access_token:
            _, data, __ = self._access_token.split(".")
            # Add additional padding to overcome the library issue when it's stripped
            data = json.loads(base64.urlsafe_b64decode(data + "==="))
            expiration = datetime.fromtimestamp(data["exp"], tz=UTC)
            refreshing_time = datetime.now(tz=UTC) + timedelta(
                minutes=self._refresh_before
            )
            self.__logger.debug(
                "Comparing %s with expiration %s from access token.",
                refreshing_time,
                expiration,
            )
            if refreshing_time > expiration:
                self._access_token = None

        if not self._access_token:
            self.__logger.debug("Refreshing OAuth access token.")
            with self.__limiter.call():
                self._access_token = self.__session.post(
                    self._oauth_url,
                    data={
                        "scope": self._oauth_scope,
                        "grant_type": self._oauth_grant_type,
                        "client_id": self._oauth_clientid,
                        "client_secret": self._oauth_clientsecret,
                    },
                ).json()["access_token"]
        return self._access_token


class RateLimiter:
    def __init__(self, max_call: int, period_seconds: int):
        self.max_call = max_call
        self.period = period_seconds
        self._last_reset = time.monotonic()
        self._counter = 0
        self._lock = RLock()

    @contextmanager
    def call(self):
        if not self.period:
            yield
            return

        while True:
            with self._lock:
                now = time.monotonic()
                if now - self._last_reset > self.period:
                    self._counter = 0
                    self._last_reset = now

                if self._counter < self.max_call:
                    self._counter += 1
                    yield
                    break
            time.sleep(now - self._last_reset)



from pathlib import Path

from setuptools import setup

BASE_REQUIREMENTS = [
    "intelmq",
    "importlib_metadata; python_version < '3.8'",
    # intelmqcli
    "python-termstyle>=0.1.10",
    "tabulate>=0.7.5",
    "psycopg2-binary",
    "rt<3.0.0,>=1.0.9",
    "mergedeep",
    # squelcher
    "netaddr>=0.7.14",
]

TESTS_REQUIRES = [
    "pytest",
    "requests-mock",
    "freezegun",
]


def generate_bots_entry_points():
    groups = ["collectors", "parsers", "experts", "outputs"]
    bots = []
    base_path = Path(__file__).parent / "intelmq_extensions/bots"
    bot_files = [
        bot_file
        for bot_file in Path(base_path).glob("**/*.py")
        if bot_file.is_file() and not bot_file.name.startswith("_")
    ]
    for file in bot_files:
        file = Path(str(file).replace(str(base_path), "intelmq_extensions/bots"))
        module = ".".join(file.with_suffix("").parts)
        bots.append("{0} = {0}:BOT.run".format(module))
        # Incoming IntelMQ extensions entrypoints
        entry_point = None
        for group in groups:
            module_prefix = "intelmq_extensions.bots.{}".format(group)
            if module.startswith(module_prefix):
                entry_point = module.replace(
                    module_prefix, "intelmq.bots.{}.certat".format(group)
                )
        if not entry_point:
            raise ValueError("Unrecognized bot type in {}".format(module))
        bots.append("{0} = {1}:BOT.run".format(entry_point, module))
    return bots


# FIXME: doesn't work for wheel build
def collect_requirements():
    requires = set(BASE_REQUIREMENTS)
    base_path = Path(__file__).parent / "intelmq_extensions/bots"
    bot_requires = [
        bot_file for bot_file in Path(base_path).glob("**/REQUIREMENTS.txt")
    ]

    for file in bot_requires:
        with open(file, "r") as f:
            requires.update(line for line in f.readlines())

    return list(requires)


setup(
    entry_points={
        "console_scripts": [
            "intelmqcli = intelmq_extensions.cli.intelmqcli:main",
            "intelmqcli_create_reports = intelmq_extensions.cli.create_reports:main",
        ]
        + generate_bots_entry_points()
    },
    install_requires=collect_requirements(),
)



"""Parser for BlackKite feeds

SPDX-FileCopyrightText: 2023 CERT.at GmbH <https://cert.at/>
SPDX-License-Identifier: AGPL-3.0-or-later
"""

import json
from dataclasses import dataclass
from typing import Union

import intelmq.lib.message as message
from intelmq.lib import utils
from intelmq.lib.bot import ParserBot

from intelmq_extensions.lib.blackkite import Category

from ._transformers import TRANSFORMERS_MAPPER


@dataclass
class PreparedData:
    raw_report: str
    company: dict
    event_data: dict


class BlackKiteParserBot(ParserBot):
    def init(self):
        pass

    def parse(self, report: message.Report):
        raw_report = utils.base64_decode(report.get("raw"))
        self._current_line = raw_report

        data = json.loads(raw_report)
        company = data["company"]
        finding = data["finding"]

        category = Category(finding["ControlId"].split("-")[0])
        for event_data in TRANSFORMERS_MAPPER[category].to_events_data(finding):
            self._current_line = PreparedData(raw_report, company, event_data)
            yield self._current_line

    def parse_line(self, data: PreparedData, report: message.Report):
        event = self.new_event(report)
        event.add("raw", data.raw_report)

        event.add("extra.monitored_asset", data.company.get("DomainName"))
        event.add("extra.blackkite_company_id", data.company.get("CompanyId"))

        for key, value in data.event_data.items():
            event.add(key, value, overwrite=True)

        if "time.source" not in event:
            event.add("time.source", event.get("time.observation"))

        return event

    def recover_line(self, line: Union[str, None, PreparedData] = None) -> str:
        if isinstance(line, str):
            return super().recover_line(line)
        return super().recover_line(line.raw_report)


BOT = BlackKiteParserBot



"""Event Group Splitter can split events based on tag groups,
and automatically recognize new groups.

The bot is designed with an intention to handle ShadowServer Compromise Website report.

SPDX-FileCopyrightText: 2024 CERT.at GmbH <https://cert.at/>
SPDX-License-Identifier: AGPL-3.0-or-later
"""

import itertools
import json
import re

from intelmq.lib.bot import ExpertBot


class EventGroupSplitterExpertBot(ExpertBot):
    look_in: str = "extra.tag"
    copy_to: list[str] = ["classification.identifier", "extra.vulnerabilities"]
    regex: str = ""  # has to match one group meaning matched duplication value
    ignore: list[str] = []  # tags to ignore
    # path to a JSON file holding groups; the file will be updated with new groups
    # the structure is a dict with keys as feed codes, and values as list of list of tags
    # e.g.:
    # {
    #   "feed-1": [["t1", "t2"], ["t1", "t3"]],
    #   "feed-2": [["t6"]]
    # }
    groups_file: str = None
    # whether groups should be analysed separately for every feed or not;
    # if not, the bot will not use "new group" path for events matching
    # group already seen in any other feed
    treat_feeds_separately: bool = True

    def init(self):
        self.matcher = re.compile(self.regex)

        self._tag_groups: dict[str, list[list[str]]] = []
        self._reload_groups()

    def _reload_groups(self, dump_current=False):
        if dump_current:
            with open(self.groups_file, "w+") as f:
                json.dump(
                    {
                        feed: [list(g) for g in groups]
                        for feed, groups in self._tag_groups.items()
                    },
                    f,
                    indent=4,
                )

        with open(self.groups_file) as f:
            new_groups = json.load(f)

        self._tag_groups = {
            feed: sorted(
                (set(group) for group in groups), key=lambda g: len(g), reverse=True
            )
            for feed, groups in new_groups.items()
        }

    def process(self):
        event = self.receive_message()

        lookup_data = event.get(self.look_in, "")
        matches = self.matcher.findall(lookup_data)
        if not matches:
            self.send_message(event)
        else:
            matches = [m for m in matches if m not in self.ignore]
            matches_set = set(matches)
            feed = event.get("feed.code", "")
            if self.treat_feeds_separately:
                feed_groups = self._tag_groups.get(feed, [])
            else:
                feed_groups = itertools.chain(*self._tag_groups.values())

            for group in feed_groups:
                if not matches_set >= group:
                    # The group is not in event's tags
                    continue
                self._generate_event(event, group)

                # Tags can be duplicated in event. Respect it.
                for tag in group:
                    matches.remove(tag)
                matches_set = set(matches)

                if not matches:
                    break

            # Something is left - new group
            if matches:
                new_group = set(matches)
                self.logger.info("New tag group was discovered: %s.", new_group)
                self._generate_event(event, new_group, new_group=True)

                if feed not in self._tag_groups:
                    self._tag_groups[feed] = list()
                self._tag_groups[feed].append(new_group)
                self._reload_groups(dump_current=True)

        self.acknowledge_message()

    def _generate_event(self, event, group: set, new_group: bool = False):
        grouped_value = "-".join(sorted(group))
        self.logger.debug("Found: %s.", group)
        sub_event = self.new_event(event)
        for key in self.copy_to:
            sub_event.add(key, grouped_value, overwrite=True)
        self.send_message(sub_event)
        if new_group:
            self.send_message(sub_event, path="new_tag_groups", path_permissive=True)


BOT = EventGroupSplitterExpertBot



"""Collector of data from BlackKite API

SPDX-FileCopyrightText: 2023 CERT.at GmbH <https://cert.at/>
SPDX-License-Identifier: AGPL-3.0-or-later

Parameters:

url
client_id
client_secret
refresh_before

categories (dict):
    {category-code, e.g. PATCH}:
        severities: [{list of names}] (optional, override)
        outputs: [{list of names}] (optional, override)
        statuses: [{list of names}] (optional, override)
        include: [{list of ids, eg. XXX-001}] (optional, mutual exclusive with exclude)
        exclude: [{list of ids, eg. XXX-001}] (optional, mutual exclusive with include)
        acknowledge: bool (optional, whether change finding's status to acknowledged or not
                     (default: false))
severities: []{list of names}
outputs: [{list of names}]
statuses: [{list of names}]
"""

import json

from intelmq.lib.bot import CollectorBot
from intelmq.lib.utils import create_request_session

from intelmq_extensions.lib.blackkite import Category

from ._client import CATEGORIES_WITH_OUTPUT, BlackKiteClient, Output, Severity, Status


class BlackKiteCollectorBot(CollectorBot):
    url: str = ""
    client_id: str = ""
    client_secret: str = ""
    # refresh access token when it's less than 10 minutes to expire
    refresh_token_before: int = 10
    # BlackKite has API rate limit 60 req./1 minute
    limit_requests: int = 60
    limit_period: int = 60  # 1 minute
    page_size: int = 100

    categories: dict = {}
    severities: list = [Severity.CRITICAL.value]
    outputs: list = [Output.FAILED.value]
    statuses: list = [Status.ACTIVE.value]

    def init(self):
        self.set_request_parameters()
        self.session = create_request_session(self)
        self.client = BlackKiteClient(
            url=self.url,
            client_id=self.client_id,
            client_secret=self.client_secret,
            refresh_before=self.refresh_token_before,
            session=self.session,
            logger=self.logger,
            limit_requests=self.limit_requests,
            limit_period=self.limit_period,
            page_size=self.page_size,
        )

        self._process_settings()

    def _process_settings(self):
        self._default_config = {
            "severities": [Severity(value) for value in self.severities],
            "outputs": [Output(value) for value in self.outputs],
            "statuses": [Status(value) for value in self.statuses],
        }

        self._categories = dict(
            self._process_category_config(k, v) for k, v in self.categories.items()
        )

    @staticmethod
    def _process_category_config(key: str, data: dict):
        data = data or {}
        category = Category(key)
        if "include" in data and "exclude" in data:
            raise ValueError("Including and excluding at the same time isn't possible")
        processed_data = {}
        if severities := data.get("severities"):
            processed_data["severities"] = [Severity(v) for v in severities]
        if statuses := data.get("statuses"):
            processed_data["statuses"] = [Status(v) for v in statuses]
        if outputs := data.get("outputs"):
            if category not in CATEGORIES_WITH_OUTPUT:
                raise ValueError(f"{category.value} does not support output filtering.")
            processed_data["outputs"] = [Output(v) for v in outputs]
        if include := data.get("include"):
            if any(filter(lambda v: not v.startswith(f"{category.value}-"), include)):
                raise ValueError(f"Category {category.value} includes incorrect IDs")
            processed_data["include"] = include
        if exclude := data.get("exclude"):
            if any(filter(lambda v: not v.startswith(f"{category.value}-"), exclude)):
                raise ValueError(f"Category {category.value} excludes incorrect IDs")
            processed_data["exclude"] = exclude
        processed_data["acknowledge"] = data.get("acknowledge", False)

        return category, processed_data

    def process(self):
        for company in self.client.companies():
            for category, config in self._categories.items():
                try:
                    self._process_category(
                        company, category, {**self._default_config, **config}
                    )
                except Exception as exc:
                    self.logger.error(
                        "Error when processing a category: %s.", exc, exc_info=True
                    )

    def _process_category(self, company: dict, category: Category, config: dict):
        def _should_send(finding):
            if include := config.get("include"):
                return finding.get("ControlId") in include
            elif exclude := config.get("exclude"):
                return finding.get("ControlId") not in exclude

            return True

        for finding in self.client.get_findings_from_category(
            category,
            company["CompanyId"],
            statuses=config.get("statuses"),
            severities=config.get("severities"),
            outputs=config.get("outputs"),
        ):
            if not _should_send(finding):
                continue
            report = self.new_report()
            report.add("raw", json.dumps({"company": company, "finding": finding}))
            self.send_message(report)
            if config.get("acknowledge"):
                self.client.acknowledge_finding(
                    company.get("CompanyId"), finding.get("FindingId")
                )

    @staticmethod
    def check(parameters: dict) -> list[list[str]] or None:
        errors = []
        definitions = [
            ("severities", Severity),
            ("outputs", Output),
            ("statuses", Status),
        ]
        for key, type_ in definitions:
            try:
                [type_(value) for value in parameters.get(key, [])]
            except ValueError as exc:
                errors.append(["error", f"Error processing '{key}': {exc}."])

        allowed_keys = {
            "severities",
            "outputs",
            "statuses",
            "include",
            "exclude",
            "acknowledge",
        }
        for category, value in parameters.get("categories", {}).items():
            value = value or {}
            try:
                if set(value.keys()) - allowed_keys:
                    raise ValueError("Unsupported config key")
                BlackKiteCollectorBot._process_category_config(category, value)
            except (ValueError, TypeError) as exc:
                errors.append(
                    ["error", f"Error processing category {category}: {exc}."]
                )

        return errors or None


BOT = BlackKiteCollectorBot



from copy import deepcopy

from intelmq.lib.bot import OutputBot
from intelmq.lib.utils import create_request_session


class MattermostOutputBot(OutputBot):
    mm_url: str
    bot_token: str
    channel_id: str

    message: str = None

    # https://developers.mattermost.com/integrate/reference/message-attachments/
    fallback: str = None
    pretext: str = None
    text: str = None

    title: str = None
    title_link: str = None

    author_name: str = "IntelMQ"
    author_icon: str = None
    author_link: str = None

    color: str = None

    fields: list[dict] = None

    footer: str = None

    # https://developers.mattermost.com/integrate/webhooks/incoming/#parameters
    card: str = None

    _template_fields = [
        "fallback",
        "message",
        "pretext",
        "text",
        "title",
        "value",
        "footer",
    ]
    _attachment_fields = [
        "fallback",
        "pretext",
        "text",
        "title",
        "title_link",
        "author_name",
        "author_icon",
        "author_link",
        "color",
        "fields",
        "footer",
    ]
    _is_attachment = False

    def init(self):
        self.set_request_parameters()
        self.session = create_request_session(self)

        if not self.message and not self.text:
            raise ValueError("Either message or text have to be configured")

        if any(getattr(self, f, None) for f in self._attachment_fields):
            self._is_attachment = True

    def process(self):
        event = self.receive_message()
        event.set_default_value()

        request_data = {"channel_id": self.channel_id, "props": {}}

        if self.message:
            request_data["message"] = self.message.format(ev=event)

        if self._is_attachment:
            request_data["props"]["attachments"] = [self._prepare_attachment(event)]

        if self.card:
            request_data["props"]["card"] = self.card.format(ev=event)

        result = self.session.post(
            f"{self.mm_url}/api/v4/posts",
            json=request_data,
            headers={"Authorization": f"Bearer {self.bot_token}"},
        )
        result.raise_for_status()

        self.acknowledge_message()

    def _prepare_attachment(self, event) -> dict:
        attachment = {}
        for field in self._attachment_fields:
            data = getattr(self, field, None)
            if data is None:
                continue
            if field == "fields":
                data: list[dict] = deepcopy(data)
                for item in data:
                    if "title" in item:
                        item["title"] = item["title"].format(ev=event)
                    if "value" in item:
                        item["value"] = item["value"].format(ev=event)
            elif field in self._template_fields:
                data = data.format(ev=event)

            attachment[field] = data
        return attachment


BOT = MattermostOutputBot
